@@ -13,22 +13,23 @@ from __future__ import division
 # compatibility functions and utilities
 from ._utils import _supports_unicode, _environ_cols_wrapper, _range, _unich, \
     _term_move_up, _unicode, WeakSet, _basestring, _OrderedDict
+from ._monitor import TMonitor
 # native libraries
 import sys
 from numbers import Number
-from threading import Thread
 from time import time
-from time import sleep
 from contextlib import contextmanager
 # For parallelism safety
 import multiprocessing as mp
 import threading as th
-
+from warnings import warn
 
 __author__ = {"github.com/": ["noamraph", "obiwanus", "kmike", "hadim",
                               "casperdcl", "lrq3000"]}
 __all__ = ['tqdm', 'trange',
-           'TqdmTypeError', 'TqdmKeyError', 'TqdmDeprecationWarning']
+           'TqdmTypeError', 'TqdmKeyError', 'TqdmWarning',
+           'TqdmExperimentalWarning', 'TqdmDeprecationWarning',
+           'TqdmMonitorWarning']
 
 
 class TqdmTypeError(TypeError):
@@ -39,19 +40,43 @@ class TqdmKeyError(KeyError):
     pass
 
 
-class TqdmDeprecationWarning(Exception):
-    # not suppressed if raised
+class TqdmWarning(Warning):
+    """base class for all tqdm warnings.
+
+    Used for non-external-code-breaking errors, such as garbled printing.
+    """
     def __init__(self, msg, fp_write=None, *a, **k):
         if fp_write is not None:
-            fp_write("\nTqdmDeprecationWarning: " + str(msg).rstrip() + '\n')
+            fp_write("\n" + self.__class__.__name__ + ": " + str(msg).rstrip() + '\n')
         else:
-            super(TqdmDeprecationWarning, self).__init__(msg, *a, **k)
+            super(TqdmWarning, self).__init__(msg, *a, **k)
+
+class TqdmExperimentalWarning(TqdmWarning, FutureWarning):
+    """beta feature, unstable API and behaviour"""
+    pass
+
+class TqdmDeprecationWarning(TqdmWarning, DeprecationWarning):
+    # not suppressed if raised
+    pass
+
+
+class TqdmMonitorWarning(TqdmWarning, RuntimeWarning):
+    """tqdm monitor errors which do not affect external functionality"""
+    pass
 
 
 # Create global parallelism locks to avoid racing issues with parallel bars
 # works only if fork available (Linux, MacOSX, but not on Windows)
-mp_lock = mp.Lock()  # multiprocessing lock
-th_lock = th.Lock()  # thread lock
+try:
+    mp_lock = mp.RLock()  # multiprocessing lock
+except ImportError:  # pragma: no cover
+    mp_lock = None
+except OSError:  # pragma: no cover
+    mp_lock = None
+try:
+    th_lock = th.RLock()  # thread lock
+except OSError:  # pragma: no cover
+    th_lock = None
 
 
 class TqdmDefaultWriteLock(object):
@@ -63,7 +88,7 @@ class TqdmDefaultWriteLock(object):
     """
     def __init__(self):
         global mp_lock, th_lock
-        self.locks = [mp_lock, th_lock]
+        self.locks = [lk for lk in [mp_lock, th_lock] if lk is not None]
 
     def acquire(self):
         for lock in self.locks:
@@ -80,80 +105,6 @@ class TqdmDefaultWriteLock(object):
         self.release()
 
 
-class TMonitor(Thread):
-    """
-    Monitoring thread for tqdm bars.
-    Monitors if tqdm bars are taking too much time to display
-    and readjusts miniters automatically if necessary.
-
-    Parameters
-    ----------
-    tqdm_cls  : class
-        tqdm class to use (can be core tqdm or a submodule).
-    sleep_interval  : fload
-        Time to sleep between monitoring checks.
-    """
-
-    # internal vars for unit testing
-    _time = None
-    _sleep = None
-
-    def __init__(self, tqdm_cls, sleep_interval):
-        Thread.__init__(self)
-        self.daemon = True  # kill thread when main killed (KeyboardInterrupt)
-        self.was_killed = False
-        self.woken = 0  # last time woken up, to sync with monitor
-        self.tqdm_cls = tqdm_cls
-        self.sleep_interval = sleep_interval
-        if TMonitor._time is not None:
-            self._time = TMonitor._time
-        else:
-            self._time = time
-        if TMonitor._sleep is not None:
-            self._sleep = TMonitor._sleep
-        else:
-            self._sleep = sleep
-        self.start()
-
-    def exit(self):
-        self.was_killed = True
-        # self.join()  # DO NOT, blocking event, slows down tqdm at closing
-        return self.report()
-
-    def run(self):
-        cur_t = self._time()
-        while True:
-            # After processing and before sleeping, notify that we woke
-            # Need to be done just before sleeping
-            self.woken = cur_t
-            # Sleep some time...
-            self._sleep(self.sleep_interval)
-            # Quit if killed
-            # if self.exit_event.is_set():  # TODO: should work but does not...
-            if self.was_killed:
-                return
-            # Then monitor!
-            # Acquire lock (to access _instances)
-            with self.tqdm_cls.get_lock():
-                cur_t = self._time()
-                # Check tqdm instances are waiting too long to print
-                for instance in self.tqdm_cls._instances:
-                    # Only if mininterval > 1 (else iterations are just slow)
-                    # and last refresh exceeded maxinterval
-                    if instance.miniters > 1 and \
-                            (cur_t - instance.last_print_t) >= \
-                            instance.maxinterval:
-                        # force bypassing miniters on next iteration
-                        # (dynamic_miniters adjusts mininterval automatically)
-                        instance.miniters = 1
-                        # Refresh now! (works only for manual tqdm)
-                        instance.refresh(nolock=True)
-
-    def report(self):
-        # return self.is_alive()  # TODO: does not work...
-        return not self.was_killed
-
-
 class tqdm(object):
     """
     Decorate an iterable object, returning an iterator which acts exactly
@@ -185,7 +136,7 @@ class tqdm(object):
         out  : str
             Number with Order of Magnitude SI unit postfix.
         """
-        for unit in ['', 'K', 'M', 'G', 'T', 'P', 'E', 'Z']:
+        for unit in ['', 'k', 'M', 'G', 'T', 'P', 'E', 'Z']:
             if abs(num) < 999.95:
                 if abs(num) < 99.95:
                     if abs(num) < 9.995:
@@ -272,7 +223,7 @@ class tqdm(object):
             The iteration unit [default: 'it'].
         unit_scale  : bool or int or float, optional
             If 1 or True, the number of iterations will be printed with an
-            appropriate SI metric prefix (K = 10^3, M = 10^6, etc.)
+            appropriate SI metric prefix (k = 10^3, M = 10^6, etc.)
             [default: False]. If any other non-zero number, will scale
             `total` and `n`.
         rate  : float, optional
@@ -289,10 +240,12 @@ class tqdm(object):
               rate_inv, rate_inv_fmt, elapsed, remaining, desc, postfix.
             Note that a trailing ": " is automatically removed after {desc}
             if the latter is empty.
-        postfix  : str, optional
+        postfix  : *, optional
             Similar to `prefix`, but placed at the end
             (e.g. for additional stats).
-            Note: postfix is a string for this method. Not a dict.
+            Note: postfix is usually a string (not a dict) for this method,
+            and will if possible be set to postfix = ', ' + postfix.
+            However other types are supported (#382).
         unit_divisor  : float, optional
             [default: 1000], ignored unless `unit_scale` is True.
 
@@ -336,6 +289,11 @@ class tqdm(object):
             n_fmt = str(n)
             total_fmt = str(total)
 
+        try:
+            postfix = ', ' + postfix if postfix else ''
+        except TypeError:
+            pass
+
         # total is known: we can predict some stats
         if total:
             # fractional and percentage progress
@@ -354,8 +312,7 @@ class tqdm(object):
                 l_bar = ''
             l_bar += '{0:3.0f}%|'.format(percentage)
             r_bar = '| {0}/{1} [{2}<{3}, {4}{5}]'.format(
-                n_fmt, total_fmt, elapsed_str, remaining_str, rate_fmt,
-                ', ' + postfix if postfix else '')
+                n_fmt, total_fmt, elapsed_str, remaining_str, rate_fmt, postfix)
 
             if ncols == 0:
                 return l_bar[:-1] + r_bar[1:]
@@ -380,7 +337,7 @@ class tqdm(object):
                             'l_bar': l_bar,
                             'r_bar': r_bar,
                             'desc': prefix or '',
-                            'postfix': ', ' + postfix if postfix else '',
+                            'postfix': postfix,
                             # 'bar': full_bar  # replaced by procedure below
                             }
 
@@ -435,8 +392,7 @@ class tqdm(object):
         else:
             return ((prefix + ": ") if prefix else '') + \
                 '{0}{1} [{2}, {3}{4}]'.format(
-                    n_fmt, unit, elapsed_str, rate_fmt,
-                    ', ' + postfix if postfix else '')
+                    n_fmt, unit, elapsed_str, rate_fmt, postfix)
 
     def __new__(cls, *args, **kwargs):
         # Create a new instance
@@ -444,25 +400,29 @@ class tqdm(object):
         # Add to the list of instances
         if "_instances" not in cls.__dict__:
             cls._instances = WeakSet()
+        if "_lock" not in cls.__dict__:
+            cls._lock = TqdmDefaultWriteLock()
         with cls._lock:
             cls._instances.add(instance)
         # Create the monitoring thread
         if cls.monitor_interval and (cls.monitor is None or not
                                      cls.monitor.report()):
-            cls.monitor = TMonitor(cls, cls.monitor_interval)
+            try:
+                cls.monitor = TMonitor(cls, cls.monitor_interval)
+            except Exception as e:  # pragma: nocover
+                warn("tqdm:disabling monitor support"
+                     " (monitor_interval = 0) due to:\n" + str(e),
+                     TqdmMonitorWarning)
+                cls.monitor_interval = 0
         # Return the instance
         return instance
 
     @classmethod
     def _get_free_pos(cls, instance=None):
-        """ Skips specified instance """
-        try:
-            return max(inst.pos for inst in cls._instances
-                       if inst is not instance) + 1
-        except ValueError as e:
-            if "arg is an empty sequence" in str(e):
-                return 0
-            raise  # pragma: no cover
+        """Skips specified instance"""
+        positions = set(abs(inst.pos) for inst in cls._instances
+                        if inst is not instance)
+        return min(set(range(len(positions) + 1)).difference(positions))
 
     @classmethod
     def _decr_instances(cls, instance):
@@ -470,23 +430,27 @@ class tqdm(object):
         Remove from list and reposition other bars
         so that newer bars won't overlap previous bars
         """
-        try:  # in case instance was explicitly positioned, it won't be in set
-            with cls._lock:
+        with cls._lock:
+            try:
                 cls._instances.remove(instance)
+            except KeyError:
+                if not instance.gui:  # pragma: no cover
+                    raise
+            else:
                 for inst in cls._instances:
-                    if inst.pos > instance.pos:
+                    # negative `pos` means fixed
+                    if inst.pos > abs(instance.pos):
                         inst.pos -= 1
-            # Kill monitor if no instances are left
-            if not cls._instances and cls.monitor:
-                try:
-                    cls.monitor.exit()
-                    del cls.monitor
-                except AttributeError:  # pragma: nocover
-                    pass
-                else:
-                    cls.monitor = None
-        except KeyError:
-            pass
+                        # TODO: check this doesn't overwrite another fixed bar
+        # Kill monitor if no instances are left
+        if not cls._instances and cls.monitor:
+            try:
+                cls.monitor.exit()
+                del cls.monitor
+            except AttributeError:  # pragma: nocover
+                pass
+            else:
+                cls.monitor = None
 
     @classmethod
     def write(cls, s, file=None, end="\n", nolock=False):
@@ -528,7 +492,6 @@ class tqdm(object):
                 inst.refresh(nolock=True)
         if not nolock:
             cls._lock.release()
-        # TODO: make list of all instances incl. absolutely positioned ones?
 
     @classmethod
     def set_lock(cls, lock):
@@ -591,21 +554,21 @@ class tqdm(object):
                     Data (may be grouped).
                 func  : function
                     To be applied on the (grouped) data.
-                *args, *kwargs  : optional
+                **kwargs  : optional
                     Transmitted to `df.apply()`.
                 """
+
                 # Precompute total iterations
                 total = getattr(df, 'ngroups', None)
                 if total is None:  # not grouped
-                    if isinstance(df, Series):
+                    if df_function == 'applymap':
+                        total = df.size
+                    elif isinstance(df, Series):
                         total = len(df)
-                    else:
-                        if kwargs.get('axis') == 1:
-                            total = len(df)
-                        else:
-                            total = df.size // len(df)
-                else:
-                    total += 1  # pandas calls update once too many
+                    else:  # DataFrame or Panel
+                        axis = kwargs.get('axis', 0)
+                        # when axis=0, total is shape[axis1]
+                        total = df.size // df.shape[axis]
 
                 # Init bar
                 if deprecated_t[0] is not None:
@@ -614,14 +577,27 @@ class tqdm(object):
                 else:
                     t = tclass(*targs, total=total, **tkwargs)
 
+                if len(args) > 0:
+                    # *args intentionally not supported (see #244, #299)
+                    TqdmDeprecationWarning(
+                        "Except func, normal arguments are intentionally" +
+                        " not supported by" +
+                        " `(DataFrame|Series|GroupBy).progress_apply`." +
+                        " Use keyword arguments instead.",
+                        fp_write=getattr(t.fp, 'write', sys.stderr.write))
+
                 # Define bar updating wrapper
                 def wrapper(*args, **kwargs):
-                    t.update()
+                    # update tbar correctly
+                    # it seems `pandas apply` calls `func` twice
+                    # on the first column/row to decide whether it can
+                    # take a fast or slow code path; so stop when t.total==t.n
+                    t.update(n=1 if t.total and t.n < t.total else 0)
                     return func(*args, **kwargs)
 
-                # Apply the provided function (in *args and **kwargs)
+                # Apply the provided function (in **kwargs)
                 # on the df using our wrapper (which provides bar updating)
-                result = getattr(df, df_function)(wrapper, *args, **kwargs)
+                result = getattr(df, df_function)(wrapper, **kwargs)
 
                 # Close bar and return pandas calculation result
                 t.close()
@@ -720,11 +696,13 @@ class tqdm(object):
             (current/instantaneous speed) [default: 0.3].
         bar_format  : str, optional
             Specify a custom bar string formatting. May impact performance.
-            If unspecified, will use '{l_bar}{bar}{r_bar}', where l_bar is
-            '{desc}: {percentage:3.0f}%|' and r_bar is
-            '| {n_fmt}/{total_fmt} [{elapsed}<{remaining}, {rate_fmt}]'
-            Possible vars: bar, n, n_fmt, total, total_fmt, percentage,
-            rate, rate_fmt, elapsed, remaining, l_bar, r_bar, desc.
+            [default: '{l_bar}{bar}{r_bar}'], where
+            l_bar='{desc}: {percentage:3.0f}%|' and
+            r_bar='| {n_fmt}/{total_fmt} [{elapsed}<{remaining}, '
+              '{rate_fmt}{postfix}]'
+            Possible vars: l_bar, bar, r_bar, n, n_fmt, total, total_fmt,
+              percentage, rate, rate_fmt, rate_noinv, rate_noinv_fmt,
+              rate_inv, rate_inv_fmt, elapsed, remaining, desc, postfix.
             Note that a trailing ": " is automatically removed after {desc}
             if the latter is empty.
         initial  : int, optional
@@ -734,10 +712,9 @@ class tqdm(object):
             Specify the line offset to print this bar (starting from 0)
             Automatic if unspecified.
             Useful to manage multiple bars at once (eg, from threads).
-        postfix  : dict, optional
+        postfix  : dict or *, optional
             Specify additional stats to display at the end of the bar.
-            Note: postfix is a dict ({'key': value} pairs) for this method,
-            not a string.
+            Calls `set_postfix(**postfix)` if possible (dict).
         unit_divisor  : float, optional
             [default: 1000], ignored unless `unit_scale` is True.
         gui  : bool, optional
@@ -841,7 +818,10 @@ class tqdm(object):
         self.bar_format = bar_format
         self.postfix = None
         if postfix:
-            self.set_postfix(refresh=False, **postfix)
+            try:
+                self.set_postfix(refresh=False, **postfix)
+            except TypeError:
+                self.postfix = postfix
 
         # Init the iterations counters
         self.last_print_n = initial
@@ -851,19 +831,18 @@ class tqdm(object):
         # not overwrite the outer progress bar
         if position is None:
             self.pos = self._get_free_pos(self)
-        else:
-            self.pos = position
-            self._instances.remove(self)
+        else:  # mark fixed positions as negative
+            self.pos = -position
 
         if not gui:
             # Initialize the screen printer
             self.sp = self.status_printer(self.fp)
             with self._lock:
                 if self.pos:
-                    self.moveto(self.pos)
+                    self.moveto(abs(self.pos))
                 self.sp(self.__repr__(elapsed=0))
                 if self.pos:
-                    self.moveto(-self.pos)
+                    self.moveto(-abs(self.pos))
 
         # Init the time counter
         self.last_print_t = self._time()
@@ -896,13 +875,13 @@ class tqdm(object):
             self.bar_format, self.postfix, self.unit_divisor)
 
     def __lt__(self, other):
-        return self.pos < other.pos
+        return abs(self.pos) < abs(other.pos)
 
     def __le__(self, other):
         return (self < other) or (self == other)
 
     def __eq__(self, other):
-        return self.pos == other.pos
+        return abs(self.pos) == abs(other.pos)
 
     def __ne__(self, other):
         return not (self == other)
@@ -917,7 +896,7 @@ class tqdm(object):
         return id(self)
 
     def __iter__(self):
-        ''' Backward-compatibility to use: for x in tqdm(iterable) '''
+        """Backward-compatibility to use: for x in tqdm(iterable)"""
 
         # Inlining instance variables as locals (speed optimisation)
         iterable = self.iterable
@@ -968,11 +947,11 @@ Please use `tqdm_gui(...)` instead of `tqdm(..., gui=True)`
                         self.n = n
                         with self._lock:
                             if self.pos:
-                                self.moveto(self.pos)
+                                self.moveto(abs(self.pos))
                             # Print bar update
                             sp(self.__repr__())
                             if self.pos:
-                                self.moveto(-self.pos)
+                                self.moveto(-abs(self.pos))
 
                         # If no `miniters` was specified, adjust automatically
                         # to the max iteration rate seen so far between 2 prints
@@ -1057,13 +1036,13 @@ Please use `tqdm_gui(...)` instead of `tqdm(..., gui=True)`
 
                 with self._lock:
                     if self.pos:
-                        self.moveto(self.pos)
+                        self.moveto(abs(self.pos))
 
                     # Print bar update
                     self.sp(self.__repr__())
 
                     if self.pos:
-                        self.moveto(-self.pos)
+                        self.moveto(-abs(self.pos))
 
                 # If no `miniters` was specified, adjust automatically to the
                 # maximum iteration rate seen so far between two prints.
@@ -1102,7 +1081,7 @@ Please use `tqdm_gui(...)` instead of `tqdm(..., gui=True)`
         self.disable = True
 
         # decrement instance pos and remove from internal set
-        pos = self.pos
+        pos = abs(self.pos)
         self._decr_instances(self)
 
         # GUI mode
@@ -1211,8 +1190,9 @@ Please use `tqdm_gui(...)` instead of `tqdm(..., gui=True)`
 
     def moveto(self, n):
         self.fp.write(_unicode('\n' * n + _term_move_up() * -n))
+        self.fp.flush()
 
-    def clear(self, nomove=False, nolock=False):
+    def clear(self, nolock=False):
         """
         Clear current bar display
         """
@@ -1221,14 +1201,10 @@ Please use `tqdm_gui(...)` instead of `tqdm(..., gui=True)`
 
         if not nolock:
             self._lock.acquire()
-        if not nomove:
-            self.moveto(self.pos)
-        # clear up the bar (can't rely on sp(''))
-        self.fp.write('\r')
-        self.fp.write(' ' * (self.ncols if self.ncols else 10))
+        self.moveto(abs(self.pos))
+        self.sp('')
         self.fp.write('\r')  # place cursor back at the beginning of line
-        if not nomove:
-            self.moveto(-self.pos)
+        self.moveto(-abs(self.pos))
         if not nolock:
             self._lock.release()
 
@@ -1241,12 +1217,9 @@ Please use `tqdm_gui(...)` instead of `tqdm(..., gui=True)`
 
         if not nolock:
             self._lock.acquire()
-        self.moveto(self.pos)
-        # clear up this line's content (whatever there was)
-        self.clear(nomove=True, nolock=True)
-        # Print current/last bar state
-        self.fp.write(self.__repr__())
-        self.moveto(-self.pos)
+        self.moveto(abs(self.pos))
+        self.sp(self.__repr__())
+        self.moveto(-abs(self.pos))
         if not nolock:
             self._lock.release()
 
